"""
Wavelet functions.
"""

## This code is written by Davide Albanese, <albanese@fbk.eu>.
## (C) 2008 Fondazione Bruno Kessler - Via Santa Croce 77, 38100 Trento, ITALY.

## See: Practical Guide to Wavelet Analysis - C. Torrence and G. P. Compo.

## This program is free software: you can redistribute it and/or modify
## it under the terms of the GNU General Public License as published by
## the Free Software Foundation, either version 3 of the License, or
## (at your option) any later version.

## This program is distributed in the hope that it will be useful,
## but WITHOUT ANY WARRANTY; without even the implied warranty of
## MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
## GNU General Public License for more details.

## You should have received a copy of the GNU General Public License
## along with this program.  If not, see <http://www.gnu.org/licenses/>.


from numpy import *
import cwb as waveletb # wb for pure python functions


__all__ = ["angularfreq", "scales", 'compute_s0', "morlet", "paul", "dog"]


def angularfreq(N, dt):
    """Compute angular frequencies.

    Input
    
      * *N*  - [integer] number of data samples
      * *dt* - [float] time step

    Output
    
      * *angular frequencies* - [1D numpy array float]

    """

    # See (5) at page 64.
    
    N2 = N / 2.0
    w = empty(N)

    for i in range(w.shape[0]):       
        if i <= N2:
            w[i] = (2 * pi * i) / (N * dt)
        else:
            w[i] = (2 * pi * (i - N)) / (N * dt)

    return w


def scales(N, dj, dt, s0):
    """Compute scales.

    Input
    
      * *N*  - [integer] number of data samples
      * *dj* - [float] scale resolution
      * *dt* - [float] time step

    Output
    
      * *scales* - [1D numpy array float]
    """

    #  See (9) and (10) at page 67.

    J = floor(dj**-1 * log2((N * dt) / s0))
    s = empty(J + 1)
    
    for i in range(s.shape[0]):
        s[i] = s0 * 2**(i * dj)
    
    return s


def compute_s0(dt, p, wf):
    """Compute s0.
    
    Input
    
      * *dt* - [float] time step
      * *p*  - [float] omega0 ('morlet') or order ('paul', 'dog')
      * *wf* - [string] wavelet function ('morlet', 'paul', 'dog')

    Output
    
      * *s0* - [float]
    """
    
    if wf == "dog":
        return (dt * sqrt(p + 0.5)) / pi
    elif wf == "paul":
        return (dt * ((2 * p) + 1)) / (2 * pi)
    elif wf == "morlet":
        return (dt * (p + sqrt(2 + p**2))) / (2 * pi)
    else:
        raise ValueError("wavelet '%s' is not available" % wf)


def morlet(N, Np, dt, dj, w0, norm = True):
    """Fourier transformed morlet function.

    Computes internally scales and angular frequencies.

    Input
      * *N*    - [integer] number of data samples (before zero padding)
      * *Np*   - [integer] number of data samples (after zero padding)
      * *dt*   - [float] time step
      * *dj*   - [float] scale resolution (smaller values of dj give finer resolution)
      * *w0*   - [float] omega0 (frequency)
      * *norm* - [bool] normalization (True or False)

    Output
      * *(normalized) fourier transformed morlet function, scales* - 
        ((ang freq x scales) [2D numpy array complex], [1D numpy array float])
    """

    w = angularfreq(Np, dt)
    s0 = compute_s0(dt, w0, 'morlet')
    s = scales(N, dj, dt, s0)    

    return waveletb.morletft(s, w, w0, dt, norm = norm), s


def paul(N, Np, dt, dj, order, norm = True):
    """Fourier transformed paul function.

    Computes internally scales and angular frequencies.

    Input
      * *N*     - [integer] number of data samples (before zero padding)
      * *Np*    - [integer] number of data samples (after zero padding)
      * *dt*    - [float] time step
      * *dj*    - [float] scale resolution (smaller values of dj give finer resolution)
      * *order* - [float] wavelet order
      * *norm*  - [bool] normalization (True or False)

    Output
      * *(normalized) fourier transformed poul function, scales* - 
        ((ang freq x scales) [2D numpy array complex], [1D numpy array float])
    """

    w = angularfreq(Np, dt)
    s0 = compute_s0(dt, order, 'paul')
    s = scales(N, dj, dt, s0)
        
    return waveletb.paulft(s, w, order, dt, norm = norm), s
    

def dog(N, Np, dt, dj, order, norm = True):
    """Fourier transformed DOG function.
    
    Computes internally scales and angular frequencies.

    Input
      * *N*     - [integer] number of data samples (before zero padding)
      * *Np*    - [integer] number of data samples (after zero padding)
      * *dt*    - [float] time step
      * *dj*    - [float] scale resolution (smaller values of dj give finer resolution)
      * *order* - [float] wavelet order
      * *norm*  - [bool] normalization (True or False)

    Output
      * *(normalized) fourier transformed DOG function, scales* - 
        ((ang freq x scales) [2D numpy array complex], [1D numpy array float])
    """

    w = angularfreq(Np, dt)
    s0 = compute_s0(dt, order, 'dog')
    s = scales(N, dj, dt, s0)
    
    return waveletb.dogft(s, w, order, dt, norm = norm), s
